// Copyright(c) 2015-present, Gabi Melman & spdlog contributors.
// Distributed under the MIT License (http://opensource.org/licenses/MIT)

#pragma once

#include <spdlog/common.h>
#include <spdlog/details/file_helper.h>
#include <spdlog/details/null_mutex.h>
#include <spdlog/fmt/fmt.h>
#include <spdlog/sinks/base_sink.h>
#include <spdlog/details/os.h>
#include <spdlog/details/circular_q.h>
#include <spdlog/details/synchronous_factory.h>

#include <chrono>
#include <cstdio>
#include <ctime>
#include <mutex>
#include <string>

extern std::array<const char *, 7> spdlog::details::days;
extern std::array<const char *, 7> spdlog::details::full_days;
extern const std::array<const char *, 12> spdlog::details::months;
extern const std::array<const char *, 12> spdlog::details::full_months;

namespace spdlog {
namespace sinks {

/*
 * Generator of daily log file names in format basename.YYYY-MM-DD.ext
 */
struct daily_filename_calculator
{
    // Create filename for the form basename.YYYY-MM-DD
    static filename_t calc_filename(const filename_t &filename, const tm &now_tm)
    {
        filename_t basename, ext;
        std::tie(basename, ext) = details::file_helper::split_by_extension(filename);
        return fmt::format(
            SPDLOG_FILENAME_T("{}_{:04d}-{:02d}-{:02d}{}"), basename, now_tm.tm_year + 1900, now_tm.tm_mon + 1, now_tm.tm_mday, ext);
    }

    // Create formatting filename for the form basename.
    static filename_t calc_filename_fmt(const filename_t &filename, const tm &now_tm)
    {
        filename_t dest = "";
        int total_minutes;
        for (size_t i = 0; i < filename.size (); ++i)
        {
            if (filename [i] == '%') {
                if (i == filename.size () - 1)
                    throw_spdlog_ex("daily_file_sink: Invalid format");
                i += 1;
                char ch = filename [i];
                switch (ch)
                {
                case 'a':
                    dest += spdlog::details::days[static_cast<size_t>(now_tm.tm_wday)];
                    break;
                case 'A':
                    dest += spdlog::details::full_days[static_cast<size_t>(now_tm.tm_wday)];
                    break;
                case 'b':
                    dest += spdlog::details::months[static_cast<size_t>(now_tm.tm_mon)];
                    break;
                case 'B':
                    dest += spdlog::details::full_months[static_cast<size_t>(now_tm.tm_mon)];
                    break;
                case 'c':
                    dest += calc_filename_fmt("%a %b %d %H:%M:%S %Y", now_tm);
                    break;
                case 'C':
                    dest += fmt::format ("{}", (now_tm.tm_year + 1900) % 100);
                    break;
                case 'Y':
                    dest += fmt::format ("{}", now_tm.tm_year + 1900);
                    break;
                case 'D':
                case 'x':
                    dest += calc_filename_fmt("%m %d %C", now_tm);
                    break;
                case 'm':
                    dest += fmt::format ("{:02}", now_tm.tm_mon + 1);
                    break;
                case 'd':
                    dest += fmt::format ("{:02}", now_tm.tm_mday);
                    break;
                case 'H':
                    dest += fmt::format ("{:02}", now_tm.tm_hour);
                    break;
                case 'I':
                    dest += fmt::format ("{:02}", (now_tm.tm_hour % 12) == 0 ? 12 : (now_tm.tm_hour % 12));
                    break;
                case 'M':
                    dest += fmt::format ("{:02}", now_tm.tm_min);
                    break;
                case 'S':
                    dest += fmt::format ("{:02}", now_tm.tm_sec);
                    break;
                case 'p':
                    dest += (now_tm.tm_hour >= 1 && now_tm.tm_hour <= 12 ? "AM" : "PM");
                    break;
                case 'r':
                    dest += calc_filename_fmt("%I:%M:%S %p", now_tm);
                    break;
                case 'R':
                    dest += calc_filename_fmt("%H:%M", now_tm);
                    break;
                case 'T':
                case 'X':
                    dest += calc_filename_fmt("%H:%M:%S", now_tm);
                    break;
                case 'z':
                    total_minutes = spdlog::details::os::utc_minutes_offset ();
                    if (total_minutes >= 0) {
                        dest += fmt::format ("+{:02}:{:02}", total_minutes / 60, total_minutes % 60);
                    } else {
                        total_minutes = -total_minutes;
                        dest += fmt::format ("-{:02}:{:02}", total_minutes / 60, total_minutes % 60);
                    }
                    break;
                case '%':
                    dest += '%';
                    break;
                default:
                    throw_spdlog_ex("daily_file_sink: Invalid format");
                }
            } else {
                dest += filename [i];
            }
        }
        return dest;
    }
};

/*
 * Rotating file sink based on date.
 * If truncate != false , the created file will be truncated.
 * If max_files > 0, retain only the last max_files and delete previous.
 */
template<typename Mutex, typename FileNameCalc = daily_filename_calculator>
class daily_file_sink final : public base_sink<Mutex>
{
public:
    // create daily file sink which rotates on given time
    daily_file_sink(filename_t base_filename, int rotation_hour, int rotation_minute, bool truncate = false, uint16_t max_files = 0, bool is_fmt = false)
        : base_filename_(std::move(base_filename))
        , rotation_h_(rotation_hour)
        , rotation_m_(rotation_minute)
        , truncate_(truncate)
        , max_files_(max_files)
        , filenames_q_()
        , is_fmt_(is_fmt)
    {
        if (rotation_hour < 0 || rotation_hour > 23 || rotation_minute < 0 || rotation_minute > 59)
        {
            throw_spdlog_ex("daily_file_sink: Invalid rotation time in ctor");
        }

        auto now = log_clock::now();
        auto filename = is_fmt_ ? FileNameCalc::calc_filename_fmt(base_filename_, now_tm(now))
                                : FileNameCalc::calc_filename(base_filename_, now_tm(now));
        file_helper_.open(filename, truncate_);
        rotation_tp_ = next_rotation_tp_();

        if (max_files_ > 0)
        {
            init_filenames_q_();
        }
    }

    filename_t filename()
    {
        std::lock_guard<Mutex> lock(base_sink<Mutex>::mutex_);
        return file_helper_.filename();
    }

protected:
    void sink_it_(const details::log_msg &msg) override
    {
        auto time = msg.time;
        bool should_rotate = time >= rotation_tp_;
        if (should_rotate)
        {
            auto filename = is_fmt_ ? FileNameCalc::calc_filename_fmt(base_filename_, now_tm(time))
                                    : FileNameCalc::calc_filename(base_filename_, now_tm(time));
            file_helper_.open(filename, truncate_);
            rotation_tp_ = next_rotation_tp_();
        }
        memory_buf_t formatted;
        base_sink<Mutex>::formatter_->format(msg, formatted);
        file_helper_.write(formatted);

        // Do the cleaning only at the end because it might throw on failure.
        if (should_rotate && max_files_ > 0)
        {
            delete_old_();
        }
    }

    void flush_() override
    {
        file_helper_.flush();
    }

private:
    void init_filenames_q_()
    {
        using details::os::path_exists;

        filenames_q_ = details::circular_q<filename_t>(static_cast<size_t>(max_files_));
        std::vector<filename_t> filenames;
        auto now = log_clock::now();
        while (filenames.size() < max_files_)
        {
            auto filename = is_fmt_ ? FileNameCalc::calc_filename_fmt(base_filename_, now_tm(now))
                                    : FileNameCalc::calc_filename(base_filename_, now_tm(now));
            if (!path_exists(filename))
            {
                break;
            }
            filenames.emplace_back(filename);
            now -= std::chrono::hours(24);
        }
        for (auto iter = filenames.rbegin(); iter != filenames.rend(); ++iter)
        {
            filenames_q_.push_back(std::move(*iter));
        }
    }

    tm now_tm(log_clock::time_point tp)
    {
        time_t tnow = log_clock::to_time_t(tp);
        return spdlog::details::os::localtime(tnow);
    }

    log_clock::time_point next_rotation_tp_()
    {
        auto now = log_clock::now();
        tm date = now_tm(now);
        date.tm_hour = rotation_h_;
        date.tm_min = rotation_m_;
        date.tm_sec = 0;
        auto rotation_time = log_clock::from_time_t(std::mktime(&date));
        if (rotation_time > now)
        {
            return rotation_time;
        }
        return {rotation_time + std::chrono::hours(24)};
    }

    // Delete the file N rotations ago.
    // Throw spdlog_ex on failure to delete the old file.
    void delete_old_()
    {
        using details::os::filename_to_str;
        using details::os::remove_if_exists;

        filename_t current_file = file_helper_.filename();
        if (filenames_q_.full())
        {
            auto old_filename = std::move(filenames_q_.front());
            filenames_q_.pop_front();
            bool ok = remove_if_exists(old_filename) == 0;
            if (!ok)
            {
                filenames_q_.push_back(std::move(current_file));
                throw_spdlog_ex("Failed removing daily file " + filename_to_str(old_filename), errno);
            }
        }
        filenames_q_.push_back(std::move(current_file));
    }

    filename_t base_filename_;
    int rotation_h_;
    int rotation_m_;
    log_clock::time_point rotation_tp_;
    details::file_helper file_helper_;
    bool truncate_;
    uint16_t max_files_;
    details::circular_q<filename_t> filenames_q_;
    bool is_fmt_;
};

using daily_file_sink_mt = daily_file_sink<std::mutex>;
using daily_file_sink_st = daily_file_sink<details::null_mutex>;

} // namespace sinks

//
// factory functions
//
template<typename Factory = spdlog::synchronous_factory>
inline std::shared_ptr<logger> daily_logger_mt(
    const std::string &logger_name, const filename_t &filename, int hour = 0, int minute = 0, bool truncate = false, uint16_t max_files = 0)
{
    return Factory::template create<sinks::daily_file_sink_mt>(logger_name, filename, hour, minute, truncate, max_files, false);
}

template<typename Factory = spdlog::synchronous_factory>
inline std::shared_ptr<logger> daily_logger_format_mt(
    const std::string &logger_name, const filename_t &filename, int hour = 0, int minute = 0, bool truncate = false, uint16_t max_files = 0)
{
    return Factory::template create<sinks::daily_file_sink_mt>(logger_name, filename, hour, minute, truncate, max_files, true);
}

template<typename Factory = spdlog::synchronous_factory>
inline std::shared_ptr<logger> daily_logger_st(
    const std::string &logger_name, const filename_t &filename, int hour = 0, int minute = 0, bool truncate = false, uint16_t max_files = 0)
{
    return Factory::template create<sinks::daily_file_sink_st>(logger_name, filename, hour, minute, truncate, max_files, false);
}

template<typename Factory = spdlog::synchronous_factory>
inline std::shared_ptr<logger> daily_logger_format_st(
    const std::string &logger_name, const filename_t &filename, int hour = 0, int minute = 0, bool truncate = false, uint16_t max_files = 0)
{
    return Factory::template create<sinks::daily_file_sink_st>(logger_name, filename, hour, minute, truncate, max_files, true);
}
} // namespace spdlog
